# Copyright 2015-2024 Bleemeo
#
# bleemeo.com an infrastructure monitoring solution in the Cloud
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from datetime import datetime, timedelta

import requests


class ConfigurationError(Exception):
    """A ConfigurationError is raised when a misconfiguration is detected."""

    pass


class APIError(Exception):
    """
    An APIError is raised when the API returns a non-successful status code.
    It serves as superclass for AuthenticationError and ThrottleError.
    """

    def __init__(self, message: str, response: requests.Response) -> None:
        if response.status_code:
            message = f"{response.status_code}: {message}"

        super().__init__(message)
        self.response = response


class BadRequestError(APIError):
    """A BadRequestError is raised when the API returns a 400 - Bad Request status code."""

    def __init__(self, response: requests.Response) -> None:
        fields = response.json()
        message = "Bad request:\n" + "\n".join(
            [f"- {field}: {' / '.join(errors)}" for field, errors in fields.items()]
        )
        super().__init__(message, response)
        self.error_fields = fields


class AuthenticationError(APIError):
    """An AuthenticationError is raised when the API returns a 401 - Unauthorized status code."""

    pass


class ResourceNotFoundError(APIError):
    """A ResourceNotFoundError is raised when the API returns a 404 - Not Found status code."""

    def __init__(self, url: str, response: requests.Response) -> None:
        super().__init__(f"Resource {url} not found", response)
        self.url = url


class ThrottleError(APIError):
    """
    A ThrottleError is raised when the API returns a 429 - Too Many Requests status code.
    It contains the delay and the deadline until a new request can safely be sent.
    If the error is not raised from the throttle-prevention mechanism, it also contains the API response.
    """

    def __init__(
        self,
        response: requests.Response | None = None,
        delay_seconds: int | None = None,
    ) -> None:
        ds = 30
        if delay_seconds is not None:
            ds = delay_seconds
        elif response is not None and "Retry-After":
            ds = int(response.headers.get("Retry-After") or 30)

        super().__init__(
            f"Throttle error: request must be retried after {ds}s.",
            response or requests.Response(),
        )
        self.delay_seconds = ds
        self.throttle_deadline = datetime.now() + timedelta(seconds=ds)

    @classmethod
    def prevent(cls, delay_seconds: int) -> ThrottleError:
        """prevent can be used to build a ThrottleError without actually sending the request to the API."""
        return cls(response=None, delay_seconds=delay_seconds)



# Copyright 2015-2024 Bleemeo
#
# bleemeo.com an infrastructure monitoring solution in the Cloud
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .client import Client
from .enums import (
    AgentType,
    ConfigItemSource,
    ConfigItemType,
    DisconnectionReason,
    GloutonDiagnostics,
    Graph,
    ReportIncluded,
    ReportPeriod,
    Status,
    TagType,
)
from .exceptions import (
    APIError,
    AuthenticationError,
    BadRequestError,
    ConfigurationError,
    ResourceNotFoundError,
    ThrottleError,
)
from .resources import Resource

__all__ = [
    "Client",
    "AgentType",
    "ConfigItemSource",
    "ConfigItemType",
    "DisconnectionReason",
    "GloutonDiagnostics",
    "Graph",
    "ReportIncluded",
    "ReportPeriod",
    "Status",
    "TagType",
    "APIError",
    "AuthenticationError",
    "BadRequestError",
    "ConfigurationError",
    "ResourceNotFoundError",
    "ThrottleError",
    "Resource",
]



# Copyright 2015-2024 Bleemeo
#
# bleemeo.com an infrastructure monitoring solution in the Cloud
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class AgentType:
    AWS_ACCOUNT = "aws_account"
    AWS_TRUSTED_ADVISOR = "aws_trusted_advisor"
    AWS_DYNAMODB = "aws_dynamodb"
    AWS_EC2 = "aws_ec2"
    AWS_ELB = "aws_elb"
    AWS_RDS = "aws_rds"
    AWS_S3 = "aws_s3"
    AGENT = "agent"
    MONITOR = "connection_check"
    SNMP = "snmp"
    K8S = "kubernetes"
    VSPHERE_CLUSTER = "vsphere_cluster"
    VSPHERE_HOST = "vsphere_host"
    VSPHERE_VM = "vsphere_vm"


class DisconnectionReason:
    CLEAN_SHUTDOWN = 1
    AGENT_TIMEOUT = 2
    AGENT_AUTO_UPGRADE = 3
    AGENT_UPGRADE = 4


class GloutonDiagnostics:
    CRASH = 0
    ON_DEMAND = 1


class Graph:
    LINE = 0
    STACK = 1
    PIE = 2
    GAUGE = 3
    AVAILABILITY_TIMELINE = 4
    NUMBER = 5
    STATUS = 6
    SNMP_STATUS = 7
    TEXT = 8
    IMAGE = 9
    HEATMAP_STATUS = 10
    BAR = 11


class ReportPeriod:
    WEEKLY = 0
    MONTHLY = 1


class ReportIncluded:
    NONE = 0
    PARTIAL = 1
    FULL = 2


class ConfigItemSource:
    UNKNOWN = 0
    DEFAULT = 1
    FILE = 2
    ENV = 3
    API = 4


class ConfigItemType:
    ANY = 0
    INT = 1
    FLOAT = 2
    BOOL = 3
    STRING = 4
    LIST_STR = 10
    LIST_INT = 11
    MAP_STR_STR = 20
    MAP_STR_INT = 21
    THRESHOLDS = 30
    SERVICES = 31
    NAME_INSTANCES = 32
    BLACKBOX_TARGETS = 33
    PROMETHEUS_TARGETS = 34
    TYPE_SNMP_TARGETS = 35
    TYPE_LOG_INPUTS = 36


class Status:
    OK = 0
    WARNING = 1
    CRITICAL = 2
    UNKNOWN = 3


class TagType:
    AUTOMATIC_API = 0
    CREATED_BY_GLOUTON = 1
    CREATED_BY_FRONTEND = 2
    AUTOMATIC_GLOUTON = 3
    AUTOMATIC_API_SERVICE = 4
    NO_TYPE = 10



# Copyright 2015-2024 Bleemeo
#
# bleemeo.com an infrastructure monitoring solution in the Cloud
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import threading
from urllib import parse

import requests

from .exceptions import APIError, AuthenticationError


class Authenticator:
    def __init__(
        self,
        api_url: str,
        session: requests.Session,
        oauth_id: str,
        oauth_secret: str | None = None,
        username: str | None = None,
        password: str | None = None,
        oauth_initial_refresh_token: str | None = None,
    ):
        self.api_url = api_url
        self.session = session
        self.oauth_client_id = oauth_id
        self.oauth_client_secret = oauth_secret
        self.username = username
        self.password = password

        self._lock = threading.Lock()
        self._current_token: str | None = None
        self._current_refresh: str | None = oauth_initial_refresh_token

    @property
    def tokens(self) -> tuple[str, str]:
        with self._lock:
            if not isinstance(self._current_token, str):
                self.__authenticate()

            return str(self._current_token), str(self._current_refresh)

    def get_token(self, force_refetch: bool = False) -> str:
        with self._lock:
            if not self._current_token or force_refetch:
                self.__authenticate()

            return str(self._current_token)

    def __authenticate(self) -> None:
        url = parse.urljoin(self.api_url, "/o/token/")

        if self._current_refresh:
            data = {
                "grant_type": "refresh_token",
                "refresh_token": self._current_refresh,
                "client_id": self.oauth_client_id,
            }

            if self.oauth_client_secret:
                data["client_secret"] = self.oauth_client_secret

            response = self.session.post(
                url,
                headers={
                    "X-Requested-With": "XMLHttpRequest",
                    "Content-Type": "application/x-www-form-urlencoded",
                },
                data=data,
                timeout=10,
            )
            if response.status_code == 200:
                response_data = response.json()
                self._current_token = response_data["access_token"]
                self._current_refresh = response_data["refresh_token"]

                return

            if not self.username:
                raise AuthenticationError(
                    "The provided initial refresh token is invalid.", response
                )

        data = {
            "grant_type": "password",
            "username": str(self.username),
            "password": self.password or "",
            "client_id": self.oauth_client_id,
        }

        if self.oauth_client_secret:
            data["client_secret"] = self.oauth_client_secret

        response = self.session.post(
            url,
            headers={
                "X-Requested-With": "XMLHttpRequest",
                "Content-Type": "application/x-www-form-urlencoded",
            },
            data=data,
            timeout=10,
        )
        if response.status_code != 200:
            raise AuthenticationError(
                f"Failed to retrieve OAuth, status={response.status_code}", response
            )

        response_data = response.json()
        self._current_token = response_data["access_token"]
        self._current_refresh = response_data["refresh_token"]

    def logout(self) -> None:
        with self._lock:
            if not self._current_refresh:
                return

            url = parse.urljoin(self.api_url, "/o/revoke_token/")
            data = {
                "token": self._current_refresh,
                "client_id": self.oauth_client_id,
                "token_type_hint": "refresh_token",
            }

            if self.oauth_client_secret:
                data["client_secret"] = self.oauth_client_secret

            response = self.session.post(
                url,
                headers={
                    "X-Requested-With": "XMLHttpRequest",
                    "Content-type": "application/x-www-form-urlencoded",
                },
                data=data,
                timeout=10,
            )
            if response.status_code != 200:
                raise APIError(
                    f"Failed to revoke token, status={response.status_code}", response
                )

            self._current_token = None
            self._current_refresh = None
